/*
 * Copyright (c) 1995-2011, Index Data
 * All rights reserved.
 * See the file LICENSE for details.
 */

package com.indexdata.torus2;

import com.indexdata.masterkey.config.MasterkeyConfiguration;
import com.indexdata.torus2.Realm.LayerType;
import com.indexdata.torus2.exception.InternalIOException;
import com.indexdata.torus2.exception.MalformedInputException;
import com.indexdata.torus2.exception.NotFoundException;
import com.indexdata.torus2.model.Facets;
import com.indexdata.torus2.model.ParentRealm;
import com.indexdata.torus2.model.RealmDefinition;
import com.indexdata.torus2.model.Record;
import com.indexdata.torus2.model.ResultSet;
import static com.indexdata.torus2.query.MatchingUtil.parseQuery;
import com.indexdata.torus2.query.QueryException;
import com.indexdata.torus2.util.DOMSerializable;
import com.indexdata.torus2.util.DateUtil;
import com.indexdata.torus2.util.URIHelper;
import com.indexdata.utils.XmlUtils;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.concurrent.ExecutorService;
import javax.servlet.ServletContext;
import javax.ws.rs.Consumes;
import javax.ws.rs.DELETE;
import javax.ws.rs.DefaultValue;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.PUT;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.Request;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.Response.ResponseBuilder;
import javax.ws.rs.core.UriInfo;
import javax.xml.transform.TransformerException;
import org.apache.log4j.Logger;
import org.apache.log4j.MDC;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;
import org.z3950.zing.cql.CQLNode;

/**
 * RESTful end-point for the Torus' Realm
 *
 * @author jakub
 */
@Path("/")
public class RealmResource {
  @Context
  private UriInfo uriInfo;
  private Logger logger = Logger.getLogger("com.indexdata.torus2");
  @Context
  private ServletContext servletContext;

  private Response buildError(int code, String msg, Exception e) {
    logger.debug("Cause for " + code, e);
    return Response.status(code).entity("Error " + code + " - " + msg).type(
      "text/plain").build();
  }
  
  private Date getStartupDate() {
    return (Date) servletContext.getAttribute("startupDate");
  }

  private String getHostName() {
    return uriInfo.getRequestUri().getHost();
  }

  private int getPort() {
    return uriInfo.getRequestUri().getPort();
  }
  
  private String getProto() {
    return uriInfo.getRequestUri().getScheme();
  }
  
  private String getHostAndPort() {
    String port = getPort() >= 0 ? ":" + Integer.toString(getPort()) : "";
    return getHostName() + port;
  }
  
  private ExecutorService getExecutorService() {
    return (ExecutorService) servletContext.getAttribute(StartupListener.EXECUTOR_SERVICE_ATTR);
  }
  
  private MasterkeyConfiguration getConfiguration() throws InternalIOException {
    try {
      return MasterkeyConfiguration.getInstance(servletContext, "torus",
          getHostName());
    } catch (IOException ex) {
      throw new InternalIOException(ex);
    }
  }

  private static Document buildDoc(String rootTag, List<? extends DOMSerializable> items,
    int start, int count) {
    Document newDoc = XmlUtils.newDoc(rootTag);
    Element root = newDoc.getDocumentElement();
    int stop = count >= 0 && (start + count) < items.size()
      ? start + count : items.size();
    root.setAttribute("total", Integer.toString(items.size()));
    root.setAttribute("start", Integer.toString(start));
    root.setAttribute("count", Integer.toString(stop-start));
    for (int i=start; i<stop; i++) {
      DOMSerializable item = items.get(i);
      item.toDOM(root);
    }
    return newDoc;
  }

  private static Node prependFacets(Document doc, Facets facets) {
    if (facets != null) {
      Node fN = doc.adoptNode(facets.toDom().getDocumentElement());
      doc.getDocumentElement().insertBefore(fN, doc.getDocumentElement().getFirstChild());
      return fN;
    }
    return null;
  }
  
  private static Node prependQuery(Document doc, CQLNode query) 
    throws InternalIOException {
    if (query != null) {
      try {
        Document qDoc = XmlUtils.parse(new StringReader(query.toXCQL()));
        Node qRoot = doc.adoptNode(qDoc.getDocumentElement());
        Node qNode = doc.createElement("xcql");
        doc.getDocumentElement().insertBefore(qNode, 
          doc.getDocumentElement().getFirstChild());
        qNode.appendChild(qRoot);
      } catch (IOException ioe) {
        throw new InternalIOException(ioe);
      } catch (SAXException sax) {
        throw new InternalIOException(sax);
      }
    }
    return null;
  }
  
  private static List<String> parseFacetNames(String csl) {
    return csl.isEmpty() ? null : Arrays.asList(csl.trim().split(",\\s*"));
  }
  
  private static Document parseUserInput(InputStream in) 
    throws MalformedInputException {
    try {
      return XmlUtils.parse(in);
    } catch (SAXException ex) {
      throw new MalformedInputException(ex.getMessage(), ex);
    } catch (IOException ex) {
      throw new MalformedInputException(ex.getMessage(), ex);
    }
  }
  
  private static void writeXml(Node tree, StringWriter sw) 
    throws InternalIOException {
    try {
      XmlUtils.serialize(tree, sw);
    } catch (TransformerException ex) {
      throw new InternalIOException(ex);
    }
  }
  
  private static URI idAsUri(String id) throws InternalIOException {
    try {
      return new URI(URIHelper.encode(id)+"/");
    } catch (UnsupportedEncodingException ex) {
      throw new InternalIOException(ex);
    } catch (URISyntaxException use) {
      throw new InternalIOException(use);
    }
  }

  @GET
  @Produces("application/xml")
  public String listRealms() {
    String logLabel = "GET "+getHostName()+"/../ ";
    int respCode = 200;
    logger.info(logLabel + "[expects no params]");
    long starttime = System.currentTimeMillis();
    MDC.put("hostname", getHostName());
    try {
      MasterkeyConfiguration cfg = getConfiguration();
      Realm realm = Realm.openRealm(null, cfg, getHostAndPort(), getProto(), getStartupDate(), getExecutorService());
      List<RealmDefinition> rdefs = realm.listAllRealms();
      Document realmsDoc = buildDoc("realms", rdefs, 0, -1);
      StringWriter sw = new StringWriter();
      writeXml(realmsDoc, sw);
      return sw.toString();
    } catch (NotFoundException nfe) {
      respCode = 404;
      throw new WebApplicationException(nfe, buildError(respCode, nfe.getMessage(), nfe));
    } catch (InternalIOException rio) {
      respCode = 500;
      throw new WebApplicationException(rio, buildError(respCode, rio.getMessage(), rio));
    } finally {
      long elapsed = System.currentTimeMillis() - starttime;
      logger.info(logLabel + "("+respCode+") took " + elapsed + " ms");
      MDC.remove("hostname");
    }
  }

  @GET
  @Path("/{realm}/")
  @Produces("application/xml")
  public String getDefinition(
    @PathParam("realm") String realmName) {
    String logLabel = "GET "+getHostName()+"/../"+realmName+"/ ";
    int respCode = 200;
    logger.info(logLabel + "[expects no params]");
    long starttime = System.currentTimeMillis();
    MDC.put("hostname", getHostName());
    try {
      MasterkeyConfiguration cfg = getConfiguration();
      Realm realm = Realm.openRealm(realmName, cfg, getHostAndPort(), getProto(), getStartupDate(), getExecutorService());
      RealmDefinition rd = realm.getDefinition();
      StringWriter sw = new StringWriter();
      writeXml(rd.toDOM(null), sw);
      return sw.toString();
    } catch (NotFoundException nfe) {
      respCode = 404;
      throw new WebApplicationException(nfe, buildError(respCode, nfe.getMessage(), nfe));
    } catch (InternalIOException rio) {
      respCode = 500;
      throw new WebApplicationException(rio, buildError(respCode, rio.getMessage(), rio));
    } finally {
      long elapsed = System.currentTimeMillis() - starttime;
      logger.info(logLabel + "("+respCode+") took " + elapsed + " ms");
      MDC.remove("hostname");
    }
  }

  @PUT
  @Consumes("application/xml")
  @Produces("application/xml")
  @Path("/{realm}/")
  public void createRealm(
    @PathParam("realm") String realmName,
    InputStream is) {
    String logLabel = "PUT "+getHostName()+"/../"+realmName+"/ ";
    int respCode = 204;
    logger.info(logLabel + "[expects no params]");
    long starttime = System.currentTimeMillis();
    MDC.put("hostname", getHostName());
    try {
      MasterkeyConfiguration cfg = getConfiguration();
      Document doc = parseUserInput(is);
      RealmDefinition rd = RealmDefinition.fromDOM(doc, realmName);
      Realm realm = Realm.createRealm(rd, cfg, getHostAndPort(), getProto(), getStartupDate(), getExecutorService());
      if (realm == null) { //already exists
        respCode = 400;
        throw new WebApplicationException(buildError(respCode, "Realm '"+realmName+" already exists", null));
      }
    } catch (InternalIOException rio) {
      respCode = 500;
      throw new WebApplicationException(rio, buildError(respCode, rio.getMessage(), rio));
    } catch (MalformedInputException mie) {
      respCode = 400;
      throw new WebApplicationException(mie, buildError(respCode, mie.getMessage(), mie));
    } finally {
      long elapsed = System.currentTimeMillis() - starttime;
      logger.info(logLabel + "("+respCode+") took " + elapsed + " ms");
      MDC.remove("hostname");
    }
  }
  
  @POST
  @Consumes("application/xml")
  @Produces("application/xml")
  @Path("/{realm}/")
  public void updateRealm(
    @PathParam("realm") String realmName,
    InputStream is) {
    String logLabel = "POST "+getHostName()+"/../"+realmName+"/ ";
    int respCode = 204;
    logger.info(logLabel + "[expects no params]");
    long starttime = System.currentTimeMillis();
    MDC.put("hostname", getHostName());
    try {
      MasterkeyConfiguration cfg = getConfiguration();
      Document doc = parseUserInput(is);
      RealmDefinition rd = RealmDefinition.fromDOM(doc, realmName);
      Realm realm = Realm.updateRealm(rd, cfg, getHostAndPort(), getProto(), getStartupDate(), getExecutorService());
      if (realm == null) { //already exists
        respCode = 400;
        throw new WebApplicationException(buildError(respCode, "Realm '"+realmName+" does not exist", null));
      }
    } catch (InternalIOException rio) {
      respCode = 500;
      throw new WebApplicationException(rio, buildError(respCode, rio.getMessage(), rio));
    } catch (MalformedInputException mie) {
      respCode = 400;
      throw new WebApplicationException(mie, buildError(respCode, mie.getMessage(), mie));
    } finally {
      long elapsed = System.currentTimeMillis() - starttime;
      logger.info(logLabel + "("+respCode+") took " + elapsed + " ms");
      MDC.remove("hostname");
    }
  }

  @DELETE
  @Consumes("application/xml")
  @Produces("application/xml")
  @Path("/{realm}/")
  public void deleteRealm(
    @PathParam("realm") String realmName) {
    String logLabel = "DELETE "+getHostName()+"/../"+realmName+"/ ";
    int respCode = 204;
    logger.info(logLabel + "[expects no params]");
    long starttime = System.currentTimeMillis();
    MDC.put("hostname", getHostName());
    try {
      MasterkeyConfiguration cfg = getConfiguration();
      Realm.destroyRealm(realmName, cfg, getHostAndPort(), getProto());
    } catch (NotFoundException nfe) {
      respCode = 404;
      throw new WebApplicationException(nfe, buildError(respCode, nfe.getMessage(), nfe));
    } catch (InternalIOException rio) {
      respCode = 500;
      throw new WebApplicationException(rio, buildError(respCode, rio.getMessage(), rio));
    } finally {
      long elapsed = System.currentTimeMillis() - starttime;
      logger.info(logLabel + "("+respCode+") took " + elapsed + " ms");
      MDC.remove("hostname");
    }
  }

  @GET
  @Path("/{realm}/merged/")
  @Produces("application/xml")
  public Response getMergedRecords(
    @Context Request req, 
    @PathParam("realm") String realmName,
    @QueryParam("layers") @DefaultValue("original,override") String layerNames,
    @QueryParam("query") @DefaultValue("") String query,
    @QueryParam("start") @DefaultValue("0") int start,
    @QueryParam("count") @DefaultValue("-1") int count,
    @QueryParam("facets") @DefaultValue("") String facets,
    @QueryParam("allowTimedRefresh") @DefaultValue("false") boolean allowTimedRefresh,
    @QueryParam("xcql") @DefaultValue("false") boolean xcql) {

    String logLabel = "GET "+getHostName()+"/../"+realmName+"/merged/ ";
    int respCode = 200;
    logger.info(logLabel + "layers='"+layerNames+"'; query='"+query+"'; "
      +"start='"+start+"'; count='"+count+"'; facets='"+facets+"'");
    if (allowTimedRefresh) logger.warn("allowTimedRefresh is deprecated and will be ignored");
    long starttime = System.currentTimeMillis();
    MDC.put("hostname", getHostName());
    try {
      MasterkeyConfiguration cfg = getConfiguration();
      Realm realm = Realm.openRealm(realmName, cfg, getHostAndPort(), getProto(), getStartupDate(), getExecutorService());
      CQLNode cql = parseQuery(query);
      ResultSet result = realm.getAllRecords(cql,
        LayerType.parse(layerNames),
        parseFacetNames(facets));
      if (result.getLastModified() != null) {
        ResponseBuilder rb = req.evaluatePreconditions(result.getLastModified());
        if (rb != null) {//not modified
          respCode = 304;
          logger.debug(logLabel+ " not modified since "+DateUtil.serialize(result.getLastModified()));
          return rb.build();
        }
      }
      //normal response
      Document recDocs = buildDoc("records", result.getRecords(), start, count);
      recDocs.getDocumentElement().setAttribute("overriden",
        Integer.toString(result.overiddenCount()));
      recDocs.getDocumentElement().setAttribute("local",
        Integer.toString(result.localCount()));
      prependFacets(recDocs, result.getFacets());
      if (xcql) prependQuery(recDocs, cql);
      StringWriter sw = new StringWriter();
      writeXml(recDocs, sw);
      return Response.ok().entity(sw.toString())
        .lastModified(result.getLastModified()).build();
    } catch (MalformedInputException mie) {
      respCode = 400;
      throw new WebApplicationException(mie, buildError(respCode, mie.getMessage(), mie));
    } catch (NotFoundException nfe) {
      respCode = 404;
      throw new WebApplicationException(nfe, buildError(respCode, nfe.getMessage(), nfe));
    } catch (InternalIOException rio) {
      respCode = 500;
      throw new WebApplicationException(rio, buildError(respCode, rio.getMessage(), rio));
    } catch (QueryException qe) {
      respCode = 400;
      throw new WebApplicationException(qe, buildError(respCode, qe.getMessage(), qe));
    } finally {
      long elapsed = System.currentTimeMillis() - starttime;
      logger.info(logLabel + "("+respCode+") took " + elapsed + " ms");
      MDC.remove("hostname");
    }
  }

  @GET
  @Path("/{realm}/records/")
  @Produces("application/xml")
  public Response getRecords(
    @Context Request req,
    @PathParam("realm") String realmName,
    @QueryParam("layers") @DefaultValue("final") String layerNames,
    @QueryParam("start") @DefaultValue("0") int start,
    @QueryParam("count") @DefaultValue("-1") int count,
    @QueryParam("query") @DefaultValue("") String query,
    @QueryParam("facets") @DefaultValue("") String facets,
    @QueryParam("allowTimedRefresh") @DefaultValue("false") boolean allowTimedRefresh,
    @QueryParam("xcql") @DefaultValue("false") boolean xcql) {

    String logLabel = "GET "+getHostName()+"/../"+realmName+"/records/ ";
    int respCode = 200;
    logger.info(logLabel + "layers='"+layerNames+"'; query='"+query+"'; "
      +"start='"+start+"'; count='"+count+"'; facets='"+facets+"'");
    if (allowTimedRefresh) logger.warn("allowTimedRefresh is deprecated and will be ignored");
    long starttime = System.currentTimeMillis();
    MDC.put("hostname", getHostName());
    try {
      MasterkeyConfiguration cfg = getConfiguration();
      Realm realm = Realm.openRealm(realmName, cfg, getHostAndPort(), getProto(), getStartupDate(), getExecutorService());
      CQLNode cql = parseQuery(query);
      ResultSet result = realm.getRealmRecords(cql,
        LayerType.parse(layerNames),
        parseFacetNames(facets));
      if (result.getLastModified() != null) {
        ResponseBuilder rb = req.evaluatePreconditions(result.getLastModified());
        if (rb != null) {//not modified
          respCode = 304;
          logger.debug(logLabel+ " not modified since "+DateUtil.serialize(result.getLastModified()));
          return rb.build();
        }
      }
      //normal response
      Document recDocs = buildDoc("records", result.getRecords(), start, count);
      prependFacets(recDocs, result.getFacets());
      if (xcql) prependQuery(recDocs, cql);
      StringWriter sw = new StringWriter();
      writeXml(recDocs, sw);
      return Response.ok().entity(sw.toString())
        .lastModified(result.getLastModified()).build();
    } catch (MalformedInputException mie) {
      respCode = 400;
      throw new WebApplicationException(mie, buildError(respCode, mie.getMessage(), mie));
    } catch (NotFoundException nfe) {
      respCode = 404;
      throw new WebApplicationException(nfe, buildError(respCode, nfe.getMessage(), nfe));
    } catch (InternalIOException rio) {
      respCode = 500;
      throw new WebApplicationException(rio, buildError(respCode, rio.getMessage(), rio));
    } catch (QueryException qe) {
      respCode = 400;
      throw new WebApplicationException(qe, buildError(respCode, qe.getMessage(), qe));
    } finally {
      long elapsed = System.currentTimeMillis() - starttime;
      logger.info(logLabel + "("+respCode+") took " + elapsed + " ms");
      MDC.remove("hostname");
    }
  }

  @GET
  @Path("/{realm}/records/{id}/")
  @Produces("application/xml")
  public Response getRecord(
    @Context Request req,
    @PathParam("realm") String realmName,
    @PathParam("id") String id,
    @QueryParam("layers") @DefaultValue("override,original,final") String layerNames,
    @QueryParam("allowTimedRefresh") @DefaultValue("false") boolean allowTimedRefresh) {

    String logLabel = "GET "+getHostName()+"/../"+realmName+"/records/"+id+"/ ";
    int respCode = 200;
    logger.info(logLabel + "layers='"+layerNames+"'");
    if (allowTimedRefresh) logger.warn("allowTimedRefresh is deprecated and will be ignored");
    long start = System.currentTimeMillis();
    MDC.put("hostname", getHostName());
    try {
      MasterkeyConfiguration cfg = getConfiguration();
      Realm realm = Realm.openRealm(realmName, cfg, getHostAndPort(), getProto(), getStartupDate(), getExecutorService());
      List<Realm.LayerType> lt = LayerType.parse(layerNames);
      Record rec = realm.getRealmRecord(id, lt);
      if (rec.getLastModified() != null) {
        ResponseBuilder rb = req.evaluatePreconditions(rec.getLastModified());
        if (rb != null) {//not modified
          respCode = 304;
          logger.debug(logLabel+ " not modified since "+DateUtil.serialize(rec.getLastModified()));
          return rb.build();
        }
      }
      //normal response
      StringWriter sw = new StringWriter();
      writeXml(rec.toDOM(null), sw);
      return Response.ok().lastModified(rec.getLastModified())
        .entity(sw.toString()).build();
    } catch (MalformedInputException mie) {
      respCode = 400;
      throw new WebApplicationException(mie, buildError(respCode, mie.getMessage(), mie));
    } catch (NotFoundException nfe) {
      respCode = 404;
      throw new WebApplicationException(nfe, buildError(respCode, nfe.getMessage(), nfe));
    } catch (InternalIOException rio) {
      respCode = 500;
      throw new WebApplicationException(rio, buildError(respCode, rio.getMessage(), rio));
    } finally {
      long elapsed = System.currentTimeMillis() - start;
      logger.info(logLabel + "("+respCode+") took " + elapsed + " ms");
      MDC.remove("hostname");
    }
  }

  @GET
  @Path("/{realm}/world/")
  @Produces("application/xml")
  public Response getWolrdRecords(
    @Context Request req,
    @PathParam("realm") String realmName,
    @QueryParam("query") @DefaultValue("") String query,
    @QueryParam("start") @DefaultValue("0") int start,
    @QueryParam("count") @DefaultValue("-1") int count,
    @QueryParam("recursive") @DefaultValue("0") int recursive,
    @QueryParam("facets") @DefaultValue("") String facets,
    @QueryParam("allowTimedRefresh") @DefaultValue("false") boolean allowTimedRefresh,
    @QueryParam("xcql") @DefaultValue("false") boolean xcql) {

    String logLabel = "GET "+getHostName()+"/../"+realmName+"/world/ ";
    int respCode = 200;
    logger.info(logLabel + "query='"+query+"'; "
      +"start='"+start+"'; count='"+count+"'; facets='"+facets+"'");
    if (allowTimedRefresh) logger.warn("allowTimedRefresh is deprecated and will be ignored");
    long starttime = System.currentTimeMillis();
    MDC.put("hostname", getHostName());
    try {
      MasterkeyConfiguration cfg = getConfiguration();
      Realm realm = Realm.openRealm(realmName, cfg, getHostAndPort(), getProto(), getStartupDate(), getExecutorService());
      CQLNode cql = parseQuery(query);
      ResultSet result = realm.getWorldRecords(cql,
        parseFacetNames(facets), recursive);
      if (result.getLastModified() != null) {
        ResponseBuilder rb = req.evaluatePreconditions(result.getLastModified());
        if (rb != null) {//not modified
          respCode = 304;
          logger.debug(logLabel+ " not modified since "+DateUtil.serialize(result.getLastModified()));
          return rb.build();
        }
      }
      //normal response
      Document recDocs = buildDoc("records", result.getRecords(), start, count);
      prependFacets(recDocs, result.getFacets());
      if (xcql) prependQuery(recDocs, cql);
      StringWriter sw = new StringWriter();
      writeXml(recDocs, sw);
      return Response.ok().entity(sw.toString())
        .lastModified(result.getLastModified()).build();
    } catch (NotFoundException nfe) {
      respCode = 404;
      throw new WebApplicationException(nfe, buildError(respCode, nfe.getMessage(), nfe));
    } catch (InternalIOException rio) {
      respCode = 500;
      throw new WebApplicationException(rio, buildError(respCode, rio.getMessage(), rio));
    } catch (QueryException qe) {
      respCode = 400;
      throw new WebApplicationException(qe, buildError(respCode, qe.getMessage(), qe));
    } finally {
      long elapsed = System.currentTimeMillis() - starttime;
      logger.info(logLabel + "("+respCode+") took " + elapsed + " ms");
      MDC.remove("hostname");
    }
  }

  @GET
  @Path("/{realm}/world/{id}/")
  @Produces("application/xml")
  public Response getWorldRecord(
    @Context Request req,
    @PathParam("realm") String realmName,
    @PathParam("id") String id,
    @QueryParam("allowTimedRefresh") @DefaultValue("false") boolean allowTimedRefresh) {

    String logLabel = "GET "+getHostName()+"/../"+realmName+"/world/"+id+"/ ";
    int respCode = 200;
    logger.info(logLabel);
    if (allowTimedRefresh) logger.warn("allowTimedRefresh is deprecated and will be ignored");
    long start = System.currentTimeMillis();
    MDC.put("hostname", getHostName());
    try {
      MasterkeyConfiguration cfg = getConfiguration();
      Realm realm = Realm.openRealm(realmName, cfg, getHostAndPort(), getProto(), getStartupDate(), getExecutorService());
      Record rec = realm.getWorldRecord(id);
      if (rec.getLastModified() != null) {
        ResponseBuilder rb = req.evaluatePreconditions(rec.getLastModified());
        if (rb != null) {//not modified
          respCode = 304;
          logger.debug(logLabel+ " not modified since "+DateUtil.serialize(rec.getLastModified()));
          return rb.build();
        }
      }
      //normal response
      StringWriter sw = new StringWriter();
      writeXml(rec.toDOM(null), sw);
      return Response.ok().lastModified(rec.getLastModified())
        .entity(sw.toString()).build();
    } catch (NotFoundException nfe) {
      respCode = 404;
      throw new WebApplicationException(nfe, buildError(respCode, nfe.getMessage(), nfe));
    } catch (InternalIOException rio) {
      respCode = 500;
      throw new WebApplicationException(rio, buildError(respCode, rio.getMessage(), rio));
    } catch (MalformedInputException mie) {
      respCode = 400;
      throw new WebApplicationException(mie, buildError(respCode, mie.getMessage(), mie));
    } finally {
      long elapsed = System.currentTimeMillis() - start;
      logger.info(logLabel + "("+respCode+") took " + elapsed + " ms");
      MDC.remove("hostname");
    }
  }

  @POST
  @Path("/{realm}/records/")
  @Consumes("application/xml")
  @Produces("application/xml")
  public Response createRecord(@PathParam("realm") String realmName,
    InputStream is) {
    String logLabel = "POST "+getHostName()+"/../"+realmName+"/records/ ";
    int respCode = 201;
    logger.info(logLabel + "[expect no params]");
    long start = System.currentTimeMillis();
    MDC.put("hostname", getHostName());
    try {
      MasterkeyConfiguration cfg = getConfiguration();
      Document doc = parseUserInput(is);
      Realm realm = Realm.openRealm(realmName, cfg, getHostAndPort(), getProto(), getStartupDate(), getExecutorService());
      Record r = Record.fromDOM(doc.getDocumentElement(), realm.getDefinition().getRealmType());
      realm.createRealmRecord(r);
      String id = r.getLayer("override").getId();
      ResponseBuilder resp = Response.created(idAsUri(id));
      return resp.build();
    } catch (MalformedInputException mre) {
      respCode = 400;
      throw new WebApplicationException(mre, buildError(respCode, mre.getMessage(), mre));
    } catch (InternalIOException rio) {
      respCode = 500;
      throw new WebApplicationException(rio, buildError(respCode, rio.getMessage(), rio));
    } finally {
      long elapsed = System.currentTimeMillis() - start;
      logger.info(logLabel + "("+respCode+") took " + elapsed + " ms");
      MDC.remove("hostname");
    }
  }

  @PUT
  @Path("/{realm}/records/{id}/")
  @Consumes("application/xml")
  @Produces("application/xml")
  public void updateRecord(@PathParam("realm") String realmName,
    @PathParam("id") String id,
    InputStream is) {
    String logLabel = "PUT "+getHostName()+"/../"+realmName+"/records/"+id+"/ ";
    int respCode = 204;
    logger.info(logLabel + "[expect no params]");
    long start = System.currentTimeMillis();
    MDC.put("hostname", getHostName());
    try {
      MasterkeyConfiguration cfg = getConfiguration();
      Document doc = parseUserInput(is);
      Realm realm = Realm.openRealm(realmName, cfg, getHostAndPort(), getProto(), getStartupDate(), getExecutorService());
      realm.updateRealmRecord(id, Record.fromDOM(doc.getDocumentElement(), realm.getDefinition().getRealmType()));
    } catch (MalformedInputException mre) {
      respCode = 400;
      throw new WebApplicationException(mre, buildError(respCode, mre.getMessage(), mre));
    } catch (NotFoundException nfe) {
      respCode = 404;
      throw new WebApplicationException(nfe, buildError(respCode, nfe.getMessage(), nfe));
    } catch (InternalIOException rio) {
      respCode = 500;
      throw new WebApplicationException(rio, buildError(respCode, rio.getMessage(), rio));
    } finally {
      long elapsed = System.currentTimeMillis() - start;
      logger.info(logLabel + "("+respCode+") took " + elapsed + " ms");
      MDC.remove("hostname");
    }
  }

  @DELETE
  @Path("/{realm}/records/{id}/")
  public void deleteRecord(@PathParam("realm") String realmName,
    @PathParam("id") String id,
    InputStream is) {
    String logLabel = "DELETE "+getHostName()+"/../"+realmName+"/records/"+id+"/ ";
    int respCode = 204;
    logger.info(logLabel + "[expect no params]");
    long start = System.currentTimeMillis();
    MDC.put("hostname", getHostName());
    try {
      MasterkeyConfiguration cfg = getConfiguration();
      Realm realm = Realm.openRealm(realmName, cfg, getHostAndPort(), getProto(), getStartupDate(), getExecutorService());
      realm.deleteRealmRecord(id);
    } catch (NotFoundException nfe) {
      respCode = 404;
      throw new WebApplicationException(nfe, buildError(respCode, nfe.getMessage(), nfe));
    } catch (InternalIOException rio) {
      respCode = 500;
      throw new WebApplicationException(rio, buildError(respCode, rio.getMessage(), rio));
    } finally {
      long elapsed = System.currentTimeMillis() - start;
      logger.info(logLabel + "("+respCode+") took " + elapsed + " ms");
      MDC.remove("hostname");
    }
  }

  @GET
  @Path("/{realm}/parents/")
  @Produces("application/xml")
  public String getParentRealms(
    @QueryParam("start") @DefaultValue("0") int start,
    @QueryParam("count") @DefaultValue("-1") int count,
    @PathParam("realm") String realmName) {

    String logLabel = "GET "+getHostName()+"/../"+realmName+"/parents/ ";
    int respCode = 200;
    logger.info(logLabel +"start='"+start+"'; count='"+count+"';");
    long starttime = System.currentTimeMillis();
    MDC.put("hostname", getHostName());
    try {
      MasterkeyConfiguration cfg = getConfiguration();
      Realm realm = Realm.openRealm(realmName, cfg, getHostAndPort(), getProto(), getStartupDate(), getExecutorService());
      List<ParentRealm> parents = realm.getParentRealms();
      Document parentsDoc = buildDoc("parents", parents, start, count);
      StringWriter sw = new StringWriter();
      writeXml(parentsDoc, sw);
      return sw.toString();
    } catch (NotFoundException nfe) {
      respCode = 404;
      throw new WebApplicationException(nfe, buildError(respCode, nfe.getMessage(), nfe));
    } catch (InternalIOException rio) {
      respCode = 500;
      throw new WebApplicationException(rio, buildError(respCode, rio.getMessage(), rio));
    } finally {
      long elapsed = System.currentTimeMillis() - starttime;
      logger.info(logLabel + "("+respCode+") took " + elapsed + " ms");
      MDC.remove("hostname");
    }
  }

  @GET
  @Path("/{realm}/parents/{id}/")
  @Produces("application/xml")
  public String getParentRealm(
    @PathParam("realm") String realmName,
    @PathParam("id") String id) {

    String logLabel = "GET "+getHostName()+"/../"+realmName+"/parents/"+id+"/ ";
    int respCode = 200;
    logger.info(logLabel +"[expect no params]");
    long starttime = System.currentTimeMillis();
    MDC.put("hostname", getHostName());
    try {
      MasterkeyConfiguration cfg = getConfiguration();
      Realm realm = Realm.openRealm(realmName, cfg, getHostAndPort(), getProto(), getStartupDate(), getExecutorService());
      ParentRealm parent = realm.getParentRealm(id);
      StringWriter sw = new StringWriter();
      writeXml(parent.toDOM(null), sw);
      return sw.toString();
    } catch (NotFoundException nfe) {
      respCode = 404;
      throw new WebApplicationException(nfe, buildError(respCode, nfe.getMessage(), nfe));
    } catch (InternalIOException rio) {
      respCode = 500;
      throw new WebApplicationException(rio, buildError(respCode, rio.getMessage(), rio));
    } finally {
      long elapsed = System.currentTimeMillis() - starttime;
      logger.info(logLabel + "("+respCode+") took " + elapsed + " ms");
      MDC.remove("hostname");
    }
  }

  @POST
  @Path("/{realm}/parents/")
  @Produces("application/xml")
  public Response createParentRealm(
    @PathParam("realm") String realmName,
    InputStream is) {

    String logLabel = "POST "+getHostName()+"/../"+realmName+"/parents/ ";
    int respCode = 201;
    logger.info(logLabel +"[expect no params]");
    long starttime = System.currentTimeMillis();
    MDC.put("hostname", getHostName());
    try {
      MasterkeyConfiguration cfg = getConfiguration();
      Realm realm = Realm.openRealm(realmName, cfg, getHostAndPort(), getProto(), getStartupDate(), getExecutorService());
      Document doc = parseUserInput(is);
      ParentRealm pr = ParentRealm.fromDOM(doc, null);
      realm.createParentRealm(pr);
      ResponseBuilder resp = Response.created(idAsUri(pr.getId()));
      return resp.build();
    } catch (RealmClient.RealmClientException rce) {
      respCode = 400;
      throw new WebApplicationException(rce, buildError(respCode, rce.getMessage(), rce));
    } catch (MalformedInputException mre) {
      respCode = 400;
      throw new WebApplicationException(mre, buildError(respCode, mre.getMessage(), mre));
    } catch (InternalIOException rio) {
      respCode = 500;
      throw new WebApplicationException(rio, buildError(respCode, rio.getMessage(), rio));
    } finally {
      long elapsed = System.currentTimeMillis() - starttime;
      logger.info(logLabel + "("+respCode+") took " + elapsed + " ms");
      MDC.remove("hostname");
    }
  }

  @PUT
  @Path("/{realm}/parents/{id}/")
  @Produces("application/xml")
  public void updateParentRealm(
    @PathParam("realm") String realmName,
    @PathParam("id") String id,
    InputStream is) {

    String logLabel = "PUT "+getHostName()+"/../"+realmName+"/parents/"+id+"/ ";
    int respCode = 200;
    logger.info(logLabel +"[expect no params]");
    long starttime = System.currentTimeMillis();
    MDC.put("hostname", getHostName());
    try {
      MasterkeyConfiguration cfg = getConfiguration();
      Realm realm = Realm.openRealm(realmName, cfg, getHostAndPort(), getProto(), getStartupDate(), getExecutorService());
      Document doc = parseUserInput(is);
      ParentRealm pr = ParentRealm.fromDOM(doc, id);
      realm.updateParentRealm(id, pr);
    } catch (NotFoundException nfe) {
      respCode = 404;
      throw new WebApplicationException(nfe, buildError(respCode, nfe.getMessage(), nfe));
    } catch (InternalIOException rio) {
      respCode = 500;
      throw new WebApplicationException(rio, buildError(respCode, rio.getMessage(), rio));
    } catch (MalformedInputException mie) {
      respCode = 400;
      throw new WebApplicationException(mie, buildError(respCode, mie.getMessage(), mie));
    } finally {
      long elapsed = System.currentTimeMillis() - starttime;
      logger.info(logLabel + "("+respCode+") took " + elapsed + " ms");
      MDC.remove("hostname");
    }
  }

  @DELETE
  @Path("/{realm}/parents/{id}/")
  public void deleteParentRealm(
    @PathParam("realm") String realmName,
    @PathParam("id") String id) {

    String logLabel = "PUT "+getHostName()+"/../"+realmName+"/parents/"+id+"/ ";
    int respCode = 200;
    logger.info(logLabel +"[expect no params]");
    long starttime = System.currentTimeMillis();
    MDC.put("hostname", getHostName());
    try {
      MasterkeyConfiguration cfg = getConfiguration();
      Realm realm = Realm.openRealm(realmName, cfg, getHostAndPort(), getProto(), getStartupDate(), getExecutorService());
      realm.deleteParentRealm(id);
    } catch (NotFoundException nfe) {
      respCode = 404;
      throw new WebApplicationException(nfe, buildError(respCode, nfe.getMessage(), nfe));
    } catch (InternalIOException rio) {
      respCode = 500;
      throw new WebApplicationException(rio, buildError(respCode, rio.getMessage(), rio));
    } finally {
      long elapsed = System.currentTimeMillis() - starttime;
      logger.info(logLabel + "("+respCode+") took " + elapsed + " ms");
      MDC.remove("hostname");
    }
  }
  
  @GET
  @Path("/{realm}/masters/")
  @Produces("application/xml")
  public String getMasterRecords(
    @PathParam("realm") String realmName,
    @QueryParam("start") @DefaultValue("0") int start,
    @QueryParam("count") @DefaultValue("-1") int count) {

    String logLabel = "GET "+getHostName()+"/../"+realmName+"/masters/ ";
    int respCode = 200;
    logger.info(logLabel +"start='"+start+"'; count='"+count+"';");
    long starttime = System.currentTimeMillis();
    MDC.put("hostname", getHostName());
    try {
      MasterkeyConfiguration cfg = getConfiguration();
      Realm realm = Realm.openRealm(realmName, cfg, getHostAndPort(), getProto(), getStartupDate(), getExecutorService());
      List<Record> masters = realm.listMasterRecords();
      Document doc = buildDoc("records", masters, start, count);
      StringWriter sw = new StringWriter();
      writeXml(doc, sw);
      return sw.toString();
    } catch (NotFoundException nfe) {
      respCode = 404;
      throw new WebApplicationException(nfe, buildError(respCode, nfe.getMessage(), nfe));
    } catch (InternalIOException rio) {
      respCode = 500;
      throw new WebApplicationException(rio, buildError(respCode, rio.getMessage(), rio));
    } finally {
      long elapsed = System.currentTimeMillis() - starttime;
      logger.info(logLabel + "("+respCode+") took " + elapsed + " ms");
      MDC.remove("hostname");
    }
  }
  
  @GET
  @Path("/{realm}/masters/{id}/")
  @Produces("application/xml")
  public String getMasterRecord(
    @PathParam("realm") String realmName,
    @PathParam("id") String id) {

    String logLabel = "GET "+getHostName()+"/../"+realmName+"/masters/"+id+"/ ";
    int respCode = 200;
    logger.info(logLabel +"[expect no params]");
    long starttime = System.currentTimeMillis();
    MDC.put("hostname", getHostName());
    try {
      MasterkeyConfiguration cfg = getConfiguration();
      Realm realm = Realm.openRealm(realmName, cfg, getHostAndPort(), getProto(), getStartupDate(), getExecutorService());
      Record master = realm.getMasterRecord(id);
      StringWriter sw = new StringWriter();
      writeXml(master.toDOM(null), sw);
      return sw.toString();
    } catch (NotFoundException nfe) {
      respCode = 404;
      throw new WebApplicationException(nfe, buildError(respCode, nfe.getMessage(), nfe));
    } catch (InternalIOException rio) {
      respCode = 500;
      throw new WebApplicationException(rio, buildError(respCode, rio.getMessage(), rio));
    } finally {
      long elapsed = System.currentTimeMillis() - starttime;
      logger.info(logLabel + "("+respCode+") took " + elapsed + " ms");
      MDC.remove("hostname");
    }
  }

  @POST
  @Path("/{realm}/masters/")
  @Produces("application/xml")
  public Response createMasterRecord(
    @PathParam("realm") String realmName,
    InputStream is) {

    String logLabel = "POST "+getHostName()+"/../"+realmName+"/parents/ ";
    int respCode = 201;
    logger.info(logLabel +"[expect no params]");
    long starttime = System.currentTimeMillis();
    MDC.put("hostname", getHostName());
    try {
      MasterkeyConfiguration cfg = getConfiguration();
      Realm realm = Realm.openRealm(realmName, cfg, getHostAndPort(), getProto(), getStartupDate(), getExecutorService());
      Document doc = parseUserInput(is);
      Record master = Record.fromDOM(doc.getDocumentElement(), "master");
      realm.createMasterRecord(master);
      ResponseBuilder resp = Response.created(idAsUri(
        master.getLayer("master-override").getId()));
      return resp.build();
    } catch (RealmClient.RealmClientException rce) {
      respCode = 400;
      throw new WebApplicationException(rce, buildError(respCode, rce.getMessage(), rce));
    } catch (MalformedInputException mre) {
      respCode = 400;
      throw new WebApplicationException(mre, buildError(respCode, mre.getMessage(), mre));
    } catch (InternalIOException rio) {
      respCode = 500;
      throw new WebApplicationException(rio, buildError(respCode, rio.getMessage(), rio));
    } finally {
      long elapsed = System.currentTimeMillis() - starttime;
      logger.info(logLabel + "("+respCode+") took " + elapsed + " ms");
      MDC.remove("hostname");
    }
  }

  @PUT
  @Path("/{realm}/masters/{id}/")
  @Produces("application/xml")
  public void updateMasterRecord(
    @PathParam("realm") String realmName,
    @PathParam("id") String id,
    InputStream is) {

    String logLabel = "PUT "+getHostName()+"/../"+realmName+"/parents/"+id+"/ ";
    int respCode = 200;
    logger.info(logLabel +"[expect no params]");
    long starttime = System.currentTimeMillis();
    MDC.put("hostname", getHostName());
    try {
      MasterkeyConfiguration cfg = getConfiguration();
      Realm realm = Realm.openRealm(realmName, cfg, getHostAndPort(), getProto(), getStartupDate(), getExecutorService());
      Document doc = parseUserInput(is);
      Record master = Record.fromDOM(doc.getDocumentElement(), "master");
      realm.updateMasterRecord(id, master);
    } catch (NotFoundException nfe) {
      respCode = 404;
      throw new WebApplicationException(nfe, buildError(respCode, nfe.getMessage(), nfe));
    } catch (InternalIOException rio) {
      respCode = 500;
      throw new WebApplicationException(rio, buildError(respCode, rio.getMessage(), rio));
    } catch (MalformedInputException mie) {
      respCode = 400;
      throw new WebApplicationException(mie, buildError(respCode, mie.getMessage(), mie));
    } finally {
      long elapsed = System.currentTimeMillis() - starttime;
      logger.info(logLabel + "("+respCode+") took " + elapsed + " ms");
      MDC.remove("hostname");
    }
  }

  @DELETE
  @Path("/{realm}/masters/{id}/")
  public void deleteMasterRecord(
    @PathParam("realm") String realmName,
    @PathParam("id") String id) {

    String logLabel = "PUT "+getHostName()+"/../"+realmName+"/parents/"+id+"/ ";
    int respCode = 200;
    logger.info(logLabel +"[expect no params]");
    long starttime = System.currentTimeMillis();
    MDC.put("hostname", getHostName());
    try {
      MasterkeyConfiguration cfg = getConfiguration();
      Realm realm = Realm.openRealm(realmName, cfg, getHostAndPort(), getProto(), getStartupDate(), getExecutorService());
      realm.deleteMasterRecord(id);
    } catch (NotFoundException nfe) {
      respCode = 404;
      throw new WebApplicationException(nfe, buildError(respCode, nfe.getMessage(), nfe));
    } catch (InternalIOException rio) {
      respCode = 500;
      throw new WebApplicationException(rio, buildError(respCode, rio.getMessage(), rio));
    } finally {
      long elapsed = System.currentTimeMillis() - starttime;
      logger.info(logLabel + "("+respCode+") took " + elapsed + " ms");
      MDC.remove("hostname");
    }
  }
}
