/*
 * Copyright (c) 1995-2011, Index Data
 * All rights reserved.
 * See the file LICENSE for details.
 */

package com.indexdata.torus2.model;

import com.indexdata.torus2.exception.MalformedInputException;
import com.indexdata.torus2.util.DOMSerializable;
import com.indexdata.torus2.model.Record.Layer.Field;
import com.indexdata.torus2.model.Record.Layer;
import java.util.LinkedList;
import java.util.List;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import static com.indexdata.torus2.util.DOMHelper.getFirstChildByName;
import static com.indexdata.torus2.util.DOMHelper.getChildrenByName;
import static com.indexdata.torus2.util.DOMHelper.createChild;

/**
 *
 * @author jakub
 */
public class RealmDefinition implements DOMSerializable {
  private String realmName;
  private String realmType;
  private String xsiLayerType;
  private boolean autoInherit;
  private MatchKey matchKey;

  public RealmDefinition(String realmName, String realmType) {
    this.realmName = realmName;
    this.realmType = realmType;
    this.xsiLayerType = realmType + "TypeLayer";
  }
  
  public RealmDefinition(String realmName, String realmType,
    boolean autoInherit, MatchKey matchKey) {
    this.realmName = realmName;
    this.realmType = realmType;
    this.xsiLayerType = realmType + "TypeLayer";
    this.autoInherit = autoInherit;
    this.matchKey = matchKey;
  }
  
  public String getRealmName() {
    return realmName;
  }

  public String getRealmType() {
    return realmType;
  }

  public MatchKey getMatchKey() {
    return matchKey;
  }

  public boolean isAutoInherit() {
    return autoInherit;
  }

  /*
   * The XsiLayerType is deduced form the realm type and it's set to ensure
   * that the
   */
  public String getXsiLayerType() {
    return xsiLayerType;
  }

  public static RealmDefinition fromDOM(Document doc, String realmName)
    throws MalformedInputException {
    Element root = doc.getDocumentElement();
    if (!"realm".equals(root.getTagName()))
      throw new MalformedInputException("Realm def should contain 'realm' as root node");
    String type = root.getAttribute("type");
    if (type == null || type.isEmpty())
      throw new MalformedInputException("Elem 'realm' - Attr 'type' missing or empty");
    boolean autoInherit = false;
    if ("yes".equalsIgnoreCase(root.getAttribute("autoInherit")))
      autoInherit = true;
    Element mKnode = getFirstChildByName(root, "matchKey");
    MatchKey matchKey = null;
    if (mKnode != null && mKnode.hasChildNodes()) {
      matchKey = new MatchKey();
      List<Element> fields = getChildrenByName(mKnode, "field");
      for (Element f : fields) {
        String fname = f.getAttribute("name");
        if (fname == null || fname.isEmpty())
          throw new MalformedInputException("Elem 'field' - Attr 'name' missing or empty");
        String freq = f.getAttribute("required");
        boolean req = false;
        if (freq != null && freq.equals("yes"))
          req = true;
        matchKey.addField(fname, req);
      }
    }
    return new RealmDefinition(realmName, type, autoInherit, matchKey);
  }

  @Override
  public Element toDOM(Element parentElement) {
    Element realmElem = createChild(parentElement, "realm");
    Document owner = realmElem.getOwnerDocument();
    realmElem.setAttribute("name", realmName);
    realmElem.setAttribute("type", realmType);
    if (autoInherit) realmElem.setAttribute("autoInherit", "yes");
    if (matchKey != null && !matchKey.fields.isEmpty()) {
      Element mK = owner.createElement("matchKey");
      for (MatchKey.MatchField mf : matchKey.fields) {
        Element mfN = owner.createElement("field");
        mfN.setAttribute("name", mf.getFieldName());
        String req = mf.isRequired() ? "yes" : "no";
        mfN.setAttribute("required", req);
        mK.appendChild(mfN);
      }
      realmElem.appendChild(mK);
    }
    return realmElem;
  }

   public static class MatchKey {
    private List<MatchField> fields = new LinkedList<MatchField>();

    public MatchField addField(String name, boolean required) {
      MatchField mf = new MatchField(name, required);
      fields.add(mf);
      return mf;
    }
    
    public String generate(Layer layer) {
      String matchKey = "";
      for (MatchField mf : fields) {
        boolean found = false;
        for (Field f : layer) {
          if (mf.fieldName.equals(f.getName())) {
            found = true;
            matchKey += f.getContents().toLowerCase();
            break;
          }
        }
        if (!found && mf.required) return layer.getId();
      }
      return matchKey.isEmpty() ? layer.getId() : matchKey;
    }

    public List<MatchField> getFields() {
      return fields;
    }

    public class MatchField {
      private String fieldName;
      private boolean required;

      public MatchField(String fieldName, boolean required) {
        this.fieldName = fieldName;
        this.required = required;
      }

      public String getFieldName() {
        return fieldName;
      }

      public boolean isRequired() {
        return required;
      }

    }

  }

}
