/*
 * Copyright (c) 1995-2011, Index Data
 * All rights reserved.
 * See the file LICENSE for details.
 */

package com.indexdata.torus2.rw;

import com.indexdata.torus2.exception.MalformedInputException;
import com.indexdata.torus2.model.ParentRealm;
import com.indexdata.torus2.model.RealmDefinition;
import com.indexdata.torus2.exception.InternalIOException;
import com.indexdata.torus2.model.Record.Layer;
import com.indexdata.torus2.util.Pair;
import java.util.Date;
import java.util.List;

/**
 * Represents low-level realm data access. General contract for this interface
 * requires that a call to either 'open' or 'create' is made BEFORE any data
 * access is attempted. Failure to fulfill this contract may result in an
 * IllegalStateException being thrown during subsequent operations, but in
 * practice may differ for specific implementations.
 * 
 * @author jakub
 */
public interface RealmReaderWriter {

  /**
   * Creates a new new realm store. Returns true if the realm store has been
   * properly created or false if the realm store already exists.
   * @throws InternalIOException when creation failed
   */
  public boolean create(RealmDefinition rd) throws InternalIOException;

  /**
   * Attempts to open the realm store pointed to by this instance. Returns true
   * if the realm has been properly opened or false if the realm has already
   * been opened (e.g on consecutive calls)
   * @return
   * @throws InternalIOException
   */
  public boolean open() throws InternalIOException;
  
  /**
   * Updates realm meta-data. Returns false when the realm does not exist.
   * @return
   * @throws InternalIOException
   */
  public boolean update(RealmDefinition rd) throws InternalIOException;

  /**
   * Retrieves RealmDefinition for the given realm.
   * @return
   * @throws InternalIOException
   */
  public RealmDefinition describe();

  /**
   * Destroys the underlying record store;
   * @throws InternalIOException
   */
  public void destroy() throws InternalIOException;

  /**
   * Retrieve a list of override records within this realm, the query
   * parameter is used ONLY to early filter out false matches, that is
   * records that CONTAIN a given query field but the value DOES NOT match.
   * @param query early filtering query, can be null
   * @return list (possibly empty e.g when realm contains no records) of records
   * @throws InternalIOException when experiencing problems reading/parsing records
   * @throws NotFoundException when records are missing RATHER than returning 'null'
   */
  public List<Layer> readOverrideRecords() throws InternalIOException;

  /**
   * Retrieve override record by id.
   * @param id override record id
   * @return found record
   * @throws InternalIOException when experiencing problems reading/parsing records
   * @throws NotFoundException when record is missing RATHER than returning 'null'
   */
  public Layer readOverrideRecord(String id) throws InternalIOException;

  /**
   * Write data to the existing record. This method does not merge anything,
   * it simply override the existing record.
   * @param id id of the records
   * @param doc new contents for the record
   * @return written contents doc, possibly with additional fields
   * @throws InternalIOException
   */
  public Layer writeOverrideRecord(String id, Layer over) throws InternalIOException;

  /**
   * The method is to write new record to the record store.
   * @param doc contents of the record to write
   * @return  written contents doc, with additional fields (e.g id)
   * @throws InternalIOException
   */
  public Layer writeOverrideRecord(Layer over) throws InternalIOException;

  public void deleteOverrideRecord(String id) throws InternalIOException;

  /**
   * Retrieve a time-stamped list of records for a parent specified within this realm,
   * @return list (possibly empty, e.g when world contains no records) of records
   * @throws InternalIOException when experiencing problems reading/parsing records
   * @throws NotFoundException when records are missing RATHER than returning 'null'
   */
  public Pair<List<Layer>,Date> readWorldRecordsForParent(String prId) throws InternalIOException;

  /**
   * Retrieve world (within a set of parents) record by id.
   * @param id override record id
   * @return found record
   * @throws InternalIOException when experiencing problems reading/parsing records
   * @throws NotFoundException when record is missing RATHER than returning 'null'
   * @throws MalformedInputException when the realmUniqueId is malformed
   */
  public Layer readWorldRecord(String realmUniqueId) throws InternalIOException, MalformedInputException;

  /**
   * Stores parent records for the given parent, updates each world record
   * with an ID unique to this realm (so world records can be looked up without
   * specifying parent)
   * @param recs
   * @param prId
   * @return
   * @throws InternalIOException
   */
  public List<Layer> writeWorldRecordsForParent(List<Layer> recs, String prId) throws InternalIOException;
  
  public void deleteWorldRecordsByParentId(String prId) throws InternalIOException;

  //----parent realm handling follows and breaks the XML convention

  public ParentRealm readParentRealm(String prId) throws InternalIOException;

  public ParentRealm writeParentRealm(String prId, ParentRealm pr) throws InternalIOException;

  public ParentRealm writeParentRealm(ParentRealm pr) throws InternalIOException;

  public void deleteParentRealm(String prId) throws InternalIOException;

  public List<ParentRealm> readParentRealms() throws InternalIOException;

  /**
   * List all realms available within the torus. This method feels welted on
   * here, but introducing tree hierarchy would complicate third-party
   * implementations.
   * @return
   */
  public List<RealmDefinition> listRealms() throws InternalIOException;

  /**
   * In case the realm serves as parent to other realms within the same torus
   * installation, the following function allows for fast propagation of changes
   * to child realms.
   */
  public boolean isCachedAsParent() throws InternalIOException;

  public Layer updateCachedAsOriginal(Layer original) throws InternalIOException;

  public void deleteCachedAsOriginal(String id) throws InternalIOException;
  
  //master record
  
  public Layer readMasterRecord(String id) throws InternalIOException;
  
  public Layer writeMasterRecord(String id, Layer record) throws InternalIOException;
  
  public Layer writeMasterRecord(Layer record) throws InternalIOException;
  
  public List<Layer> readMasterRecords() throws InternalIOException;
  
  public void deleteMasterRecord(String id) throws InternalIOException;
  
}
