/*
 * Copyright (c) 1995-2011, Index Data
 * All rights reserved.
 * See the file LICENSE for details.
 */

package com.indexdata.torus2.rw.impl;

import com.indexdata.torus2.exception.MalformedInputException;
import com.indexdata.torus2.util.Pair;
import com.indexdata.torus2.util.LRUCache;
import com.indexdata.torus2.exception.NotFoundException;
import com.indexdata.torus2.model.ParentRealm;
import com.indexdata.torus2.model.RealmDefinition;
import com.indexdata.torus2.exception.InternalIOException;
import com.indexdata.torus2.model.Record.Layer;
import com.indexdata.torus2.rw.RealmReaderWriter;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.apache.log4j.Logger;

/**
 * This RW prevents any disk writes or reads that have to do with worlds records
 * and instead caches them purely in memory. The persisted parent list is consulted
 * each time the cache is read, in case world records (full realms) are missing
 * from the cache a request is made to retrieve them.
 * @author jakub
 */
public class CachingReaderWriter extends ProxyReaderWriter {
  private String host;
  private String proto;
  private boolean honorProto;
  private boolean honorAuth;
  private boolean honorPort;
  private URL selfURL;
  private static Logger logger = Logger.getLogger("com.indexdata.torus2");
  private final static LRUCache<String,Pair<List<Layer>,Date>> cache =
    new LRUCache<String, Pair<List<Layer>,Date>>(500);
  //parents are cached per request only
  private Map<String,ParentRealm> parents;
  private RealmDefinition rdef;
  private boolean isOpen;
  
  // cache key ignores protocol and authority information
  private String key(URL url) {
    StringBuilder key = new StringBuilder();
    if (honorProto) key.append(url.getProtocol()).append("://");
    if (honorAuth) key.append(url.getAuthority()).append("@");
    key.append(url.getHost());
    if (honorPort && url.getPort() != -1) {
      key.append(":").append(url.getPort());
    }
    key.append(url.getFile());
    return key.toString();
  }

  private void readParentRealmsToMap(RealmReaderWriter rw) throws InternalIOException {
    //since the cache is shared by all realms we need the list of parents
    List<ParentRealm> prs = rw.readParentRealms();
    parents =
      new HashMap<String, ParentRealm>(prs.size());
    for (ParentRealm pr : prs) {
      parents.put(pr.getId(), pr);
    }
  }

  private Layer findRecById(List<Layer> docs, String origId) {
    for (int i=0; i<docs.size(); i++) {
      Layer candid = docs.get(i);
      if (origId.equals(candid.getId())) {
        return candid;
      }
    }
    return null;
  }

  public CachingReaderWriter(RealmReaderWriter rw, String host, String proto,
    boolean honorProto, boolean honorAuth, boolean honorPort) throws InternalIOException {
    super(rw);
    this.host = host;
    this.proto = proto;
    this.honorProto = honorProto;
    this.honorAuth = honorAuth;
    this.honorPort = honorPort;
  }

  @Override
  public boolean open() throws InternalIOException {
    isOpen = super.open();
    if (isOpen) {
      rdef = describe();
      try {
        selfURL = new URL(proto+"://"+host+"/torus2/"+rdef.getRealmName()+"/records/");
      } catch (MalformedURLException mue) {
        throw new InternalIOException("Cannot constrcut realm self URL", mue);
      }
      readParentRealmsToMap(this.rw);
    }
    return isOpen;
  }

  @Override
  public ParentRealm writeParentRealm(String prId, ParentRealm pr) throws
    InternalIOException {
    if (!isOpen) throw new IllegalStateException("Realm RW is closed");
    ParentRealm uPr = super.writeParentRealm(prId, pr);
    parents.put(prId, uPr);
    return uPr;
  }
  
  @Override
  public ParentRealm writeParentRealm(ParentRealm pr) throws InternalIOException {
    if (!isOpen) throw new IllegalStateException("Realm RW is closed");
    //super call has to be first to populate the id
    ParentRealm uPr = super.writeParentRealm(pr);
    parents.put(pr.getId(), uPr);
    return uPr;
  }

  @Override
  public void deleteParentRealm(String prId) throws InternalIOException {
    if (!isOpen) throw new IllegalStateException("Realm RW is closed");
    super.deleteParentRealm(prId);
    parents.remove(prId);
  }

  @Override
  public ParentRealm readParentRealm(String prId) throws InternalIOException {
    if (!isOpen) throw new IllegalStateException("Realm RW is closed");
    ParentRealm pr = parents.get(prId);
    if (pr == null)
      throw new NotFoundException("Missing parent for id '"+prId+"'");
    else
      return pr;
  }

  @Override
  public List<ParentRealm> readParentRealms() throws InternalIOException {
    if (!isOpen) throw new IllegalStateException("Realm RW is closed");
    return new ArrayList<ParentRealm>(parents.values());
  }

  @Override
  public Pair<List<Layer>,Date> readWorldRecordsForParent(String prId) throws InternalIOException {
    if (!isOpen) throw new IllegalStateException("Realm RW is closed");
    ParentRealm pr = parents.get(prId);
    if (pr == null) throw new NotFoundException("Parent '"+prId+"' is unknown");
    URL purl = pr.resolveUrl(host);
    logger.debug("Looking up cached parent '"+pr.getName()+"'@<"+purl+">");
    return cache.get(key(purl));
  }

  @Override
  public Layer readWorldRecord(String prefixId) throws InternalIOException, MalformedInputException {
    if (!isOpen) throw new IllegalStateException("Realm RW is closed");
    int dotidx = prefixId.indexOf('.');
    if (dotidx <= 0 || dotidx == prefixId.length())
      throw new MalformedInputException("Invalid worldId: " + prefixId);
    ParentRealm pr = parents.get(prefixId.substring(0, dotidx));
    if (pr == null) throw new NotFoundException("Parent for world record '"+prefixId+"' is unknown");
    URL purl = pr.resolveUrl(host);
    logger.debug("Looking up cached parent '"+pr.getName()+"'@<"+purl+">");
    Pair<List<Layer>,Date> rP = cache.get(key(purl));
    if (rP == null) return null;
    //cached records keep their original IDs
    Layer rec = findRecById(rP.getFirst(), prefixId.substring(dotidx+1));
    if (rec == null) throw new NotFoundException("World record '"+prefixId+"' is missing");
    return rec;
  }

  @Override
  public List<Layer> writeWorldRecordsForParent(List<Layer> docs, String prId)
    throws InternalIOException {
    if (!isOpen) throw new IllegalStateException("Realm RW is closed");
    ParentRealm pr = parents.get(prId);
    if (pr == null) throw new NotFoundException("Parent with id '"+prId+"' not found");
    URL parentUrl = pr.resolveUrl(host);
    logger.info("Caching world records for '"+pr.getName()+"'@<"+parentUrl+">");
    Pair<List<Layer>,Date> cachedRealm = cache.get(key(parentUrl));
    if (cachedRealm != null) {
      synchronized (cachedRealm) {
        cachedRealm.setFirst(docs);
        cachedRealm.setSecond(new Date());
      }
    } else {
      cachedRealm = new Pair<List<Layer>, Date>(docs, new Date());
      cache.put(key(parentUrl), cachedRealm);
    }
    return docs;
  }

  @Override
  public void deleteWorldRecordsByParentId(String prId) throws InternalIOException {
    if (!isOpen) throw new IllegalStateException("Realm RW is closed");
    ParentRealm pr = parents.get(prId);
    if (pr == null) throw new NotFoundException("Parent with id '"+prId+"' not found");
    //do nothing, old cache entries will be removed automatically
  }
  
  //following functions are overriden to ensure same-host instant world propagation

  @Override
  public boolean isCachedAsParent() throws InternalIOException {
    return cache.get(key(selfURL)) != null;
  }

  @Override
  public void deleteCachedAsOriginal(String id) throws InternalIOException {
    Pair<List<Layer>,Date> pair = cache.get(key(selfURL));
    if (pair == null)
      throw new InternalIOException("Realm '"+rdef.getRealmName()+"' is not a cached parent.");
    boolean found = false;
    List<Layer> origs = pair.getFirst();
    for (int i=0; i<origs.size(); i++) {
      Layer candid = origs.get(i);
      if (id.equals(candid.getId())) {
        found = true;
        origs.remove(i);
        break;
      }
    }
    if (!found) throw new InternalIOException("Cache inconsistent -- removed original cannot be found");
  }

  @Override
  public Layer updateCachedAsOriginal(Layer original) throws InternalIOException {
    Pair<List<Layer>,Date> pair = cache.get(key(selfURL));
    if (pair == null)
      throw new InternalIOException("Realm '"+rdef.getRealmName()+"' is not a cached parent.");
    boolean found = false;
    List<Layer> origs = pair.getFirst();
    for (int i=0; i<origs.size(); i++) {
      Layer candid = origs.get(i);
      if (original.getId().equals(candid.getId())) {
        found = true;
        origs.set(i, original);
        break;
      }
    }
    if (!found)
      origs.add(original);
    return original;
  }
}
