/*
 * Copyright (c) 1995-2011, Index Data
 * All rights reserved.
 * See the file LICENSE for details.
 */

package com.indexdata.torus2.util;

import com.indexdata.utils.XmlUtils;
import java.util.ArrayList;
import java.util.List;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Helper class for common (awkward) DOM operations on the record object.
 * May disappear if the records get an objectified wrapper.
 * @author jakub
 */
public class DOMHelper {

  /**
   * The method creates and appends Element node to the given parent. If parent
   * is null, a new document is created.
   * @param parent
   * @param name
   */
  public static Element createChild(Element parent, String name) {
    Document owner = parent != null ? parent.getOwnerDocument() : XmlUtils.newDoc();
    Element elem = owner.createElement(name);
    if (parent == null) owner.appendChild(elem); else parent.appendChild(elem);
    return elem;
  }

  /*
   * Returns the first child element of the base element with given name
   */
  public static Element getFirstChildByName(Element base, String name) {
    NodeList children = base.getChildNodes();
    for (int i=0; i<children.getLength(); i++) {
      Node child = children.item(i);
      if (child.getNodeType() == Node.ELEMENT_NODE) {
        Element elem = (Element) child;
        if (elem.getTagName().equals(name))
          return elem;
      }
    }
    return null;
  }
  
  public static String getFirstChildText(Element base, String name) {
    Element child = getFirstChildByName(base, name);
    return child != null ? child.getTextContent().trim() : null;
  }

  public static List<Element> getChildrenByName(Element base, String name) {
    List<Element> elems = new ArrayList<Element>();
    NodeList children = base.getChildNodes();
    for (int i=0; i<children.getLength(); i++) {
      Node child = children.item(i);
      if (child.getNodeType() == Node.ELEMENT_NODE) {
        Element elem = (Element) child;
        if (elem.getTagName().equals(name)) elems.add(elem);
      }
    }
    return elems;
  }

  /**
   * Set child text node within the base level, ensure there's no duplicates
   * @param base
   */
  public static void setUniqueChild(Element base, String name, String value) {
    NodeList children = base.getChildNodes();
    boolean set = false;
    for (int i = 0; i < children.getLength(); i++) {
      Node child = children.item(i);
      if (child.getNodeType() == Node.ELEMENT_NODE) {
        Element elem = (Element) child;
        if (elem.getTagName().equals(name)) {
          if (!set) {
            elem.setTextContent(value);
            set = true;
          } else {
            base.removeChild(child);
          }
        }
      }
    }
    if (!set) {
      XmlUtils.appendTextNode(base, name, value);
    }
  }

  public static Element getFirstElemWithAttr(Element base, String elemName, String attrName, String attrVal) {
    NodeList elems = base.getElementsByTagName(elemName);
    for (int i = 0; i < elems.getLength(); i++) {
      Element l = (Element) elems.item(i);
      if (attrVal.equalsIgnoreCase(l.getAttribute(attrName))) {
        return l;
      }
    }
    return null;
  }

  public static String getFirstElementText(Element base, String elemName) {
    NodeList el = base.getElementsByTagName(elemName);
    if (el != null && el.getLength() == 1) {
      return el.item(0).getTextContent().trim();
    }
    return null;
  }

}
